unit module_defs;

{

   COPYRIGHT 2013 .. 2015 Blue Hell / Jan Punter

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 2 as
  published by the Free Software Foundation;

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  For all listed email addresses :

    _dot. to be substituted by a dot      '.'
    2@t2  to be substituted by an at sign '@'


  Blue Hell is a trade mark owned by

    Jan Punter
    Oogstplein 6
    7545 HP Enschede
    the Netherlands
    http://www.iaf.nl/Users/BlueHell/
    http://bluehell.electro-music.com/
    j_dot.punter2@t2iaf_dot.nl

  All rights attributed to Blue Hell are owned by Jan Punter.
}

{.$DEFINE PATCH_PROFILER :: best to be defined from build configuration}

interface

uses

  System.SysUtils, System.Math, System.Classes,

  KnobsUtils, KnobsConversions, ComplexMath, midi_defs, OSCif;


const

  NO_INPUT      = 'noinput';
  NO_OUTPUT     = 'nooutput';
  NO_STRING     = 'nostring';
  NO_MODULE     = 'nomodule';
  UNDEF_PMC     =    -1;               // Pin/module/connection index for non existent pin/module/connection
  MAX_SIGNALS   =   256;
  MAX_INPUTS    = MAX_SIGNALS;
  MAX_OUTPUTS   = MAX_SIGNALS;
  FIFO_SIZE     = 65536;               // Sizes for input and output FIFOs, should be as large as the largest settable size for the audio device queues
  DO_DEZIPPER   = True;
  NO_DEZIPPER   = False;

  {$IFDEF PATCH_PROFILER}
  HAS_PATCH_PROFILER = True;
  {$ELSE}
  HAS_PATCH_PROFILER = False;
  {$ENDIF}

  WhiteSpace             = [ #$00 .. #$20, #$7f];
  StrictDigits           = [ '0' .. '9'];
  Digits                 = StrictDigits + [ '+', '-'];
  IdentifierStarters     = [ 'A' .. 'Z', 'a' .. 'z', '%', '#', '@', '_'];
  IdentifierChars        = IdentifierStarters + Digits + [ '&'];
  FloatChars             = Digits + [ '.', 'e', 'E'];

  inputtype_normal       = 0;
  inputtype_inverted     = 1;
  inputtype_positive     = 2;
  inputtype_pos_inv      = 3;
  inputtype_negative     = 4;
  inputtype_neg_inv      = 5;

  outputtype_normal      = 0;
  outputtype_inverted    = 1;
  outputtype_positive    = 2;
  outputtype_pos_inv     = 3;
  outputtype_negative    = 4;
  outputtype_neg_inv     = 5;

  envoutputtype_normal   = 0;
  envoutputtype_inverted = 1;
  envoutputtype_negative = 2;
  envoutputtype_neg_inv  = 3;


type

  EModule           = class( Exception);
  EModuleNotFound   = class( EModule  );
  EModuleDuplicate  = class( EModule  );
  EInputNotFound    = class( EModule  );
  EOutputNotFound   = class( EModule  );
  EStringNotFound   = class( EModule  );
  EUnitsUndefined   = class( EModule  );
  ENoPinLookup      = class( EModule  );

  TStringArray      = array of string;
  TMod              = class;
  TSynthPatch       = class;
  TModuleArray      = array of TMod;
  TModuleClass      = class of TMod;


  TModuleIdTypeMap = record
    ModuleId    : Integer;
    ModuleClass : TModuleClass;
  end;
  TModuleIdTypeMaps = array of TModuleIdTypeMap;


  TPinMap = record
    ModuleIndex : Integer;
    Ins         : TByteSet;
    Outs        : TByteSet;
  end;
  TPinMaps = array of TPinMap;


  TLightsInfo = record
  public
    Prefix : string;
    Name   : string;
    Signal : string;
    Data   : TSignal;
  end;


  TDataInfo = record
  public
    Prefix : string;
    Name   : string;
    Signal : string;
    Data   : TSignalArray;
  end;


  TXYDataInfo = record
  public
    Prefix : string;
    Name   : string;
    Signal : string;
    Data   : TSignalPairFifo;
  end;


  TInfoInfo = record
  public
    Prefix : string;
    Name   : string;
    Signal : string;
    Data   : string;
  end;


  TLightsHandler = procedure( const aSender: TMod   ; const anInfo: TLightsInfo) of object;
  TDataHandler   = procedure( const aSender: TMod   ; const anInfo: TDataInfo  ) of object;
  TXYDataHandler = procedure( const aSender: TMod   ; const anInfo: TXYDataInfo) of object;
  TInfoHandler   = procedure( const aSender: TMod   ; const anInfo: TInfoInfo  ) of object;
  TOnSynthLog    = procedure( const aSender: TObject; const aMsg : string)       of object;
  TOnSynthError  = procedure( const aSender: TObject; const aMsg : string)       of object;
  TOnOSCString   = procedure( const aSender: TObject; const aMsg: TBytes)        of object;


  TConnection = class
  public
    Peer      : TSynthPatch;
    Index     : Integer;
    SrcModule : Integer;
    SrcPin    : Integer;
    DstModule : Integer;
    DstPin    : Integer;
  protected
    procedure   Connect( const aPeer: TSynthPatch; anIndex, aSrcModule, aSrcPin, aDstModule, aDstPin: Integer);
  public
    procedure   Save( anIndent: Integer; var S: string);
    function    IsSameConnection( aConnection: TConnection): Boolean;
  end;
  TConnectionArray  = array of TConnection;


  TConnections = class
  private
    FPeer       : TSynthPatch;
    FConnections: TConnectionArray;
  private
    function    GetCount: Integer;
    function    GetConnection( anIndex: Integer): TConnection;
    procedure   SetConnection( anIndex: Integer; const aValue: TConnection);
  public
    constructor Create( aPeer: TSynthPatch);
    destructor  Destroy;                                                                                       override;
    procedure   Save( anIndent: Integer; const aName: string; var S: string);
    procedure   Clear;
    procedure   AddConnection( const aValue: TConnection);                                                     overload;
    procedure   AddConnection( aSrcModule, aSrcPin, aDstModule, aDstPin: Integer);                             overload;
    procedure   RemoveConnection( anIndex: Integer);
    function    IsSameConnections( const aConnections: TConnections): Boolean;
  public
    property    Count: Integer                             read GetCount;
    property    Connection[ anIndex: Integer]: TConnection read GetConnection write SetConnection;
  end;


  TSignals = array[ 0 .. MAX_INPUTS - 1] of PSignal;
  PSignals = ^TSignals;


  TLinks = record
  private
    FCount  : Integer;
    FSource : PSignal;
    FLinks  : PSignals;
  public
    procedure   Initialize( aSource: PSignal);
    procedure   Uninitialize;
    procedure   AddLink( aDestination: PSignal);
  end;
  TLinksArray = array[ 0 .. MAX_OUTPUTS - 1] of TLinks;
  PLinksArray = ^TLinksArray;


  TOutputLinks = class
  private
    FCount    : Integer;
    FOutLinks : PLinksArray;
  private
    function    FindSource( aSource: PSignal): Integer;
    procedure   AddLink( aSource, aDestination: PSignal);
  public
    destructor  Destroy;                                                                                       override;
    procedure   Compile( const aModule: TMod; const aPatch: TSynthPatch);
  end;


  TMod = class
  private
    FParent          : TSynthPatch;       // The patch this module is part off
    FLocked          : Boolean;           // When true module should not be executed
    FExecuting       : Boolean;           // True when the module is currently being executed
    FName            : string;            // A unique name for this module - unique within it's parent
    FResetFlag       : Integer;           // Set when a module reset was requested
    FSamplesFast     : Int64;             // Sample counter, incremented each time a new Fast sample is calcultaed
    FSamplesSlow     : Int64;             // Sample counter, incremented each time a new Slow sample is calcultaed

  {$IFDEF PATCH_PROFILER}
  private
    FClocksFast      : Int64;             // CPU Fast clocks for current cycle
    FClocksSlow      : Int64;             // CPU Slow clocks for current cycle
    FTotalClocksFast : Int64;             // CPU Fast clocks accumulated over the module life time
    FTotalClocksSlow : Int64;             // CPU Slow clocks accumulated over the module life time
    FZips            : Int64;             // CPU clocks spent on dezoppering for the current cycle
    FTotalZips       : Int64;             // CPU clocks acumulated FZips over the module life time
  {$ENDIF}

  protected
    FIsFast          : Boolean;           // True when fast mode processing must be performed
    FIsSpedUp        : Boolean;           // True when module is normally slow but sped up due to fast signals being present on an input.
    FIsSlow          : Boolean;           // True when slow mode processing must be performed
    FIsMidi          : Boolean;           // True when MIDI reception must be enabled - for MIDI modules only
    FIsOSC           : Boolean;           // True when OSC  reception must be enabled - for OSC  modules only
    FInputCount      : Integer;           // Amount of input and amount of dezipper signals
    FOutputCount     : Integer;           // Amount of output signals
    FInputs          : TSignalArray;      // The input signals, these include both patch signals and user controls
    FDezips          : TSignalArray;      // An extra set of inputs for user controls, these are copied to FInCache with some filtering when the signal happens to be be in the deziper map
    FOutputs         : TSignalArray;      // The ouput signals.
    FDezipperMap     : TByteSet;          // Determines  what signals should be dezippered, most, but not all, knobs should be dezippered
  [weak] FOutputLinks: TOutputLinks;      // For each output a set of signals that should be updated from ot, compiled connections
  private
    function    AssertUniqueName( const aName: string): string;                                                 virtual;
  private
    procedure   SetLocked( aValue: Boolean);
    procedure   SetName      ( const aValue: string);
    function    GetResetFlag: Boolean;
    procedure   SetResetFlag( aValue: Boolean);                                                                 virtual;
    function    GetOutput( anIndex: Integer): TSignal;
    procedure   SetCommonValue        ( anIndex : Integer; aValue: TSignal; Dezipper: Boolean);                 virtual;
    procedure   SetInternalValue      ( const aName: string; aValue: TSignal);                                  virtual;
    procedure   SetInternalStringValue( const aName, aValue: string);
    procedure   SetAutomationValue    ( const aName: string; aValue: TSignal);                                  virtual;
  protected
    procedure   SetStringValue        ( const aName, aValue: string);                                           virtual;
    function    GetStringValue        ( const aName: string): string;                                           virtual;
  protected
    procedure   FixInputs;
    procedure   FixOutputs;
    procedure   HandleParam          ( const aControl: string; aValue: TSignal);                                virtual;
    procedure   HandleAutomationParam( const aControl: string; aValue: TSignal);                                virtual;
    procedure   Log( const aMsg: string);                                                                       virtual;
    procedure   LogFmt( const aFmt: string; const anArgs: array of const);
    procedure   Error( const aMsg: string);                                                                     virtual;
    procedure   ErrorFmt( const aFmt: string; const anArgs: array of const);
    {$IFDEF PATCH_PROFILER}
  protected
    procedure   PreTickFast;                                                                                     inline;
    procedure   PostTickFast;                                                                                    inline;
    procedure   PreTickSlow;                                                                                     inline;
    procedure   PostTickSlow;                                                                                    inline;
    procedure   PreZipper;                                                                                       inline;
    procedure   PostZipper;                                                                                      inline;
    procedure   ClearProfile;
    {$ENDIF}
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                                             virtual;
    destructor  Destroy;                                                                                       override;
    procedure   CompileLinks;                                                                                   virtual;
    procedure   PropagateSignals;
    procedure   SetDefaults;                                                                                    virtual;
    procedure   SaveKeyValue    ( anIndent: Integer; const aName: string; aValue: TSignal; var S: string);
    procedure   SaveInputValues ( anIndent: Integer; var S: string);
    procedure   SaveOutputValues( anIndent: Integer; var S: string);
    procedure   SaveDetails     ( anIndent: Integer; var S: string);                                            virtual;
    procedure   SetInternal     ( const aName: string; aValue: TSignal);                                        virtual;
    function    GetInternal     ( const aName: string): TSignal;                                                virtual;
    procedure   Save            ( anIndent: Integer; var S: string);
    procedure   Reset;                                                                                          virtual;
    procedure   ResetSampleCounts;                                                                              virtual;
    procedure   Clear;                                                                                          virtual;
    procedure   Panic;                                                                                          virtual;
    procedure   Pulse( anInput: Integer);                                                                       virtual;
    function    FindInput   ( const aPinName: string): Integer;
    function    FindOutput  ( const aPinName: string): Integer;
    function    InputName   ( anIndex: Integer): string;                                                        virtual;
    function    OutputName  ( anIndex: Integer): string;                                                        virtual;
    procedure   MakeFast;
    procedure   Tick;
    procedure   SlowTick;
    procedure   DeZipper;
    procedure   DoTick;                                                                                         virtual;
    procedure   DoSlowTick;                                                                                     virtual;
    procedure   FixZippers;                                                                                     virtual;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);                          virtual;
    procedure   GatherData  ( const aPrefix: string; const aCallback: TDataHandler);                            virtual;
    procedure   GatherXYData( const aPrefix: string; const aCallback: TXYDataHandler);                          virtual;
    procedure   GatherInfo  ( const aPrefix: string; const aCallback: TInfoHandler);                            virtual;
    procedure   AcceptParam          ( const aPath: string; aValue: TSignal);
    procedure   AcceptAutomationParam( const aPath: string; aValue: TSignal);
    function    GetProfileInfoHtml: string;
    function    IsSameModule( const aModule: TMod): Boolean;
    procedure   AcceptMidi( const aMsg: TMidiMessage);                                                          virtual;
    procedure   SendMidi( const aMsg: TMidiMessage);                                                            virtual;
    function    OSCMatched( const aMsg: TOSCMessage): Boolean;                                                  virtual;
    procedure   AcceptOSC( const aSynthName: string; const aMsg: TOSCPacket);                                   virtual;
    procedure   SendOSCBytes( const aMsg: TBytes);                                                              virtual;
    procedure   SampleRateChanged;                                                                              virtual;
    procedure   CollectDenormals( const aRange: TSignal; var aMap: TPinMap);
  public
    property    Locked                                    : Boolean     read FLocked           write SetLocked;
    property    Name                                      : string      read FName             write SetName;
    property    ResetFlag                                 : Boolean     read GetResetFlag      write SetResetFlag;
    property    SamplesFast                               : Int64       read FSamplesFast;
    property    SamplesSlow                               : Int64       read FSamplesSlow;
    property    InputCount                                : Integer     read FInputCount;
    property    OutputCount                               : Integer     read FOutputCount;
    property    Output   [ anIndex: Integer]              : TSignal     read GetOutput;
    property    IsFast                                    : Boolean     read FIsFast;
    property    IsSlow                                    : Boolean     read FIsSlow;
    property    IsMidi                                    : Boolean     read FIsMidi;
    property    IsOSC                                     : Boolean     read FIsOSC;
    property    IsSpedUp                                  : Boolean     read FIsSpedUp;
  public
    property    InternalValue      [ const aName: string] : TSignal                            write SetInternalValue;
    property    AutomationValue    [ const aName: string] : TSignal                            write SetAutomationValue;
    property    InternalStringValue[ const aName: string] : string                             write SetInternalStringValue;
    property    StringValue        [ const aName: string] : string      read GetStringValue    write SetStringValue;
  {$IFDEF PATCH_PROFILER}
  public
    property    TotalClocksFast                           : Int64       read FTotalClocksFast;
    property    TotalClocksSlow                           : Int64       read FTotalClocksSlow;
    property    TotalZips                                 : Int64       read FTotalZips;
  {$ENDIF}
  end;


  TModInput = class( TMod)
  strict private
  const
    o_out1 = 0;
    o_out2 = 1;
    o_out3 = 2;
    o_out4 = 3;
    o_out5 = 4;
    o_out6 = 5;
    o_out7 = 6;
    o_out8 = 7;
  public
    function    OutputName( anIndex: Integer): string;                                                         override;
  end;


  TModOutput = class( TMod)
  strict private
  const
    i_in1 = 0;
    i_in2 = 1;
    i_in3 = 2;
    i_in4 = 3;
    i_in5 = 4;
    i_in6 = 5;
    i_in7 = 6;
    i_in8 = 7;
  public
    function    InputName( anIndex: Integer): string;                                                          override;
  end;


  TSynthPatch = class( TMod)
  private
    FVersion         : Integer;
    FMidiChannel     : Byte;
    FModules         : TModuleArray;
    FConnections     : TConnections;
    FDecimatorCount  : Integer;
    FInputFifo       : TSignalOctetFifo;
    FOnSendMidi      : TOnMidiMessage;
    FOnSendOscString : TOnOSCString;
    FOnSynthLog      : TOnSynthLog;
    FOnSynthError    : TOnSynthError;
  private
    function    AssertUniqueName( const aName: string): string;                                                override;
  private
    function    CheckModuleIndex( anIndex: Integer): Boolean;
    function    GetModuleCount: Integer;
    function    GetModule( anIndex: Integer): TMod;
    function    GetConnectionCount: Integer;
    procedure   SetResetFlag( aValue: Boolean);                                                                override;
    function    GetOverflows : Integer;
    function    GetUnderflows: Integer;
    function    FixPrefix( const aPrefix: string): string;
  protected
    procedure   HandleParam          ( const aPath: string; aValue: TSignal);                                  override;
    procedure   HandleAutomationParam( const aPath: string; aValue: TSignal);                                  override;
    procedure   Log( const aMsg: string);                                                                      override;
    procedure   Error( const aMsg: string);                                                                    override;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                                            override;
    destructor  Destroy;                                                                                       override;
    procedure   CompileLinks;                                                                                  override;
    procedure   SetDefaults;                                                                                   override;
    procedure   SaveConnections( anIndent: Integer; var S: string);
    procedure   SaveModules    ( anIndent: Integer; var S: string);
    procedure   SaveDetails    ( anIndent: Integer; var S: string);                                            override;
    procedure   Clear;                                                                                         override;
    procedure   ResetSampleCounts;                                                                             override;
    procedure   ClearModules;
    procedure   Panic;                                                                                         override;
    function    FindModule( const aModuleName: string): Integer;
    function    ModuleName( anIndex: Integer): string;
    function    FindInputPin ( const aModuleIndex: Integer; const aPinName: string): Integer;
    function    FindOutputPin( const aModuleIndex: Integer; const aPinName: string): Integer;
    function    InputPinName ( aModuleIndex, aPinIndex: Integer): string;
    function    OutputPinName( aModuleIndex, aPinIndex: Integer): string;
    procedure   AddModule( const aModule: TMod);
    procedure   AddConnection( aSrcModule, aSrcPin, aDstModule, aDstPin: Integer);                             overload;
    procedure   AddConnection( const aSrcModule, aSrcPin, aDstModule, aDstPin: string);                        overload;
    function    InputName ( anIndex: Integer): string;                                                         override;
    function    OutputName( anIndex: Integer): string;                                                         override;
    procedure   DoTick;                                                                                        override;
    procedure   FixZippers;                                                                                    override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);                         override;
    procedure   GatherData  ( const aPrefix: string; const aCallback: TDataHandler);                           override;
    procedure   GatherXYData( const aPrefix: string; const aCallback: TXYDataHandler);                         override;
    procedure   GatherInfo  ( const aPrefix: string; const aCallback: TInfoHandler);                           override;
    procedure   AcceptSampleOctet( aCh1, aCh2, aCh3, aCh4, aCh5, aCh6, aCh7, aCh8: TSignal);
    function    FifoFillCount: Integer;
    procedure   ClearProfile;
    function    CreateProfileHtml: TStringList;
    function    IsSamePatch    ( const aPatch: TSynthPatch): Boolean;
    function    SameModules    ( const aPatch: TSynthPatch): Boolean;
    function    SameConnections( const aPatch: TSynthPatch): Boolean;
    procedure   AcceptMidi( const aMsg: TMidiMessage);                                                         override;
    procedure   SendMidi( const aMsg: TMidiMessage);                                                           override;
    procedure   AcceptOSC( const aSynthName: string; const aMsg: TOSCPacket);                                  override;
    procedure   SendOSCBytes( const aMsg: TBytes);                                                             override;
    procedure   SampleRateChanged;                                                                             override;
    procedure   ClearOverUnderFlows;
    procedure   CollectDenormals( const aRange: TSignal; var aMap: TPinMaps);
  public
    property    Version                   : Integer        read FVersion           write FVersion;
    property    ModuleCount               : Integer        read GetModuleCount;
    property    Module[ anIndex: Integer] : TMod           read GetModule;
    property    ConnectionCount           : Integer        read GetConnectionCount;
    property    MidiChannel               : Byte           read FMidiChannel       write FMidiChannel;
    property    Overflows                 : Integer        read GetOverflows;
    property    Underflows                : Integer        read GetUnderflows;
    property    OnSendMidi                : TOnMidiMessage read FOnSendMidi        write FOnSendMidi;
    property    OnSendOSCString           : TOnOSCString   read FOnSendOscString   write FOnSendOscString;
    property    OnSynthLog                : TOnSynthLog    read FOnSynthLog        write FOnSynthLog;
    property    OnSynthError              : TOnSynthError  read FOnSynthError      write FOnSynthError;
  end;


  procedure RegisterModuleType      ( const aModuleId: Integer; aModuleClass: TModuleClass);
  function  LookupModuleClass       ( const aModuleId: Integer): TModuleClass;
  function  LookupModuleGeneralClass( const aModuleId: Integer): TClass;
  procedure ParsePrefix ( const aPath: string; var aPrefix, aRest: string);
  function  ParseTextSequencerValue( const aValue: string): TSignalArray;
  function  SignalForInputType( const aSignal: TSignal; anInputType: Integer): TSignal;                          inline;
  function  SignalForOutputType( const aSignal: TSignal; anOutputType: Integer): TSignal;                        inline;
  function  SignalForEnvOutputType( const aSignal: TSignal; anOutputType: Integer): TSignal;                     inline;
  function  PolyBLEP( aPhase, anAmount, aDeltaPhase: TSignal): TSignal;                                          inline;
  function  MakeInfo( const aPrefix, aName, aSignal: string; const aData: TSignal)        : TLightsInfo;       overload;
  function  MakeInfo( const aPrefix, aName, aSignal: string; const aData: TSignalArray)   : TDataInfo  ;       overload;
  function  MakeInfo( const aPrefix, aName, aSignal: string; const aData: TSignalPairFifo): TXYDataInfo;       overload;
  function  MakeInfo( const aPrefix, aName, aSignal: string; const aData: string)         : TInfoInfo  ;       overload;



implementation


uses

  Globals;

var

  ModuleIdTypeMaps  : TModuleIdTypeMaps;


  procedure RegisterModuleType( const aModuleId: Integer; aModuleClass: TModuleClass);
  var
    aClass : TClass;
  begin
    Assert( Assigned( aModuleClass), 'Trying to register a NIL module type');
    aClass := LookupModuleClass( aModuleId);

    if not Assigned( aClass)
    then begin
      SetLength( ModuleIdTypeMaps, Length( ModuleIdTypeMaps) + 1);
      ModuleIdTypeMaps[ Length( ModuleIdTypeMaps) - 1].ModuleId    := aModuleId;
      ModuleIdTypeMaps[ Length( ModuleIdTypeMaps) - 1].ModuleClass := aModuleClass;
    end
    else raise EModuleDuplicate.CreateFmt( 'Duplicate module ID %d for already present with type %s while trying to register type %s', [ aModuleId, aClass.ClassName, aModuleClass.ClassName]);
  end;


  function  LookupModuleClass( const aModuleId: Integer): TModuleClass;
  var
    M : TModuleIdTypeMap;
  begin
    Result := nil;

    for M in ModuleIdTypeMaps
    do begin
      if M.ModuleId = aModuleId
      then begin
        Result := M.ModuleClass;
        Break;
      end;
    end;
  end;


  function  LookupModuleGeneralClass( const aModuleId: Integer): TClass;
  var
    M : TModuleIdTypeMap;
  begin
    Result := nil;

    for M in ModuleIdTypeMaps
    do begin
      if M.ModuleId = aModuleId
      then begin
        Result := M.ModuleClass;
        Break;
      end;
    end;
  end;


  procedure ParsePrefix( const aPath: string; var aPrefix, aRest: string);
  var
    i  : Integer;
  begin
    i       := 1;
    aPrefix := '';
    aRest   := '';

    while i <= Length( aPath)
    do begin
      if aPath[ i] = '/'
      then begin
        aPrefix := Copy( aPath, 1, i - 1);
        aRest   := Copy( aPath, i + 1, Length( aPath));
        Break;
      end
      else Inc( i);
    end;
  end;



  function  ParseTextSequencerValue( const aValue: string): TSignalArray;

    procedure AddSignal( aValue: TSignal; var aResult: TSignalArray);
    var
      aLength : Integer;
    begin
      aLength := Length( aResult);
      SetLength( aResult, aLength + 1);
      aResult[ aLength] := aValue;
    end;

  var
    aStrings : TStringList;
    aString  : string;
    i        : Integer;
    j        : Integer;
    aSignal  : TSignal;
    aParts   : TStringList;
    aRepeats : Integer;
  begin
    Result := nil;
    aStrings := Explode( Trim( aValue), ',');

    try
      for i := 0 to aStrings.Count - 1
      do begin
        aString := Trim( aStrings[ i]);
        aParts  := Explode( aString, ':');

        if aParts.Count > 0
        then begin
          if aParts.Count > 1
          then aRepeats := StrToIntDef( Trim( aParts[ 1]), 1)
          else aRepeats := 1;

          if ParseTextValue( Trim( aParts[ 0]), aSignal)
          then begin
            if IsNan( aSignal)
            then aSignal := NO_NOTE;

            AddSignal( aSignal, Result);

            for j := 1 to aRepeats - 1
            do AddSignal( NO_NOTE, Result);
          end
          else AddSignal( NO_NOTE, Result);
        end
        else AddSignal( NO_NOTE, Result);
      end;
    finally
      aStrings.DisposeOf;
    end;
  end;


  function  SignalForInputType( const aSignal: TSignal; anInputType: Integer): TSignal; inline;
  // Translates input range to a [-1..1] output range
  begin
    case anInputType of
      inputtype_inverted : Result :=     - aSignal    ;         // [ 1..-1] -> [-1..1]
      inputtype_positive : Result :=   2 * aSignal - 1;         // [ 0.. 1] -> [-1..1]
      inputtype_pos_inv  : Result := - 2 * aSignal + 1;         // [ 1.. 0] -> [-1..1]
      inputtype_negative : Result :=   2 * aSignal + 1;         // [-1.. 0] -> [-1..1]
      inputtype_neg_inv  : Result := - 2 * aSignal - 1;         // [ 0..-1] -> [-1..1]
      else                 Result :=       aSignal    ;         // [-1.. 1] -> [-1..1]
    end;
  end;


  function  SignalForOutputType( const aSignal: TSignal; anOutputType: Integer): TSignal; inline;
  // Translates a [-1..1] range to an output range
  begin
    case anOutputType of
      outputtype_inverted : Result :=       - aSignal       ;   // [-1..1] -> [ 1..-1]
      outputtype_positive : Result := (   1 + aSignal) * 0.5;   // [-1..1] -> [ 0.. 1]
      outputtype_pos_inv  : Result := (   1 - aSignal) * 0.5;   // [-1..1] -> [ 1.. 0]
      outputtype_negative : Result := ( - 1 + aSignal) * 0.5;   // [-1..1] -> [-1.. 0]
      outputtype_neg_inv  : Result := ( - 1 - aSignal) * 0.5;   // [-1..1] -> [ 0..-1]
      else                  Result :=         aSignal       ;   // [-1..1] -> [-1.. 1]
    end;
  end;


  function  SignalForEnvOutputType( const aSignal: TSignal; anOutputType: Integer): TSignal; inline;
  // Translates a [0..1] range to an output range
  begin
    case anOutputType of
      envoutputtype_inverted : Result :=   - aSignal    ;       // [0..1] -> [ 0..-1]
      envoutputtype_negative : Result :=     aSignal - 1;       // [0..1] -> [-1.. 0]
      envoutputtype_neg_inv  : Result := 1 - aSignal    ;       // [0..1] -> [ 1.. 0]
      else                     Result :=     aSignal    ;       // [0..1] -> [ 0.. 1]
    end;
  end;


  function  PolyBLEP( aPhase, anAmount, aDeltaPhase: TSignal): TSignal; inline;
  // Moved out of HrastUnit.pas for usage by other oscillators than HrastOsc as well.
  var
    dt: TSignal;
  begin
    dt := anAmount * aDeltaPhase;

    if aPhase < dt
    then begin
      aPhase := aPhase / dt - 1;
      Result := - aPhase * aPhase;
    end
    else if aPhase > 1 - dt
    then begin
      aPhase := ( aPhase - 1) / dt + 1;
      Result := aPhase * aPhase;
    end
    else Result := 0
  end;


  function  MakeInfo( const aPrefix, aName, aSignal: string; const aData: TSignal): TLightsInfo; overload;
  begin
    Result.Prefix := aPrefix;
    Result.Name   := aName;
    Result.Signal := aSignal;
    Result.Data   := aData;
  end;


  function  MakeInfo( const aPrefix, aName, aSignal: string; const aData: TSignalArray): TDataInfo  ; overload;
  begin
    Result.Prefix := aPrefix;
    Result.Name   := aName;
    Result.Signal := aSignal;
    Result.Data   := aData;
  end;


  function  MakeInfo( const aPrefix, aName, aSignal: string; const aData: TSignalPairFifo): TXYDataInfo; overload;
  begin
    Result.Prefix := aPrefix;
    Result.Name   := aName;
    Result.Signal := aSignal;
    Result.Data   := aData;
  end;


  function  MakeInfo( const aPrefix, aName, aSignal: string; const aData: string): TInfoInfo  ; overload;
  begin
    Result.Prefix := aPrefix;
    Result.Name   := aName;
    Result.Signal := aSignal;
    Result.Data   := aData;
  end;


{ ========
  TConnection = class
  private
    Peer      : TSynthPatch;
    Index     : Integer;
    SrcModule : Integer;
    SrcPin    : Integer;
    DstModule : Integer;
    DstPin    : Integer;
  private
}

    procedure   TConnection.Connect( const aPeer: TSynthPatch; anIndex, aSrcModule, aSrcPin, aDstModule, aDstPin: Integer);
    begin
      Peer      := aPeer;
      Index     := anIndex;
      SrcModule := aSrcModule;
      SrcPin    := aSrcPin;
      DstModule := aDstModule;
      DstPin    := aDstPin;
    end;


//  public

    procedure   TConnection.Save( anIndent: Integer; var S: string);
    var
      aSrcModName : string;
      aSrcPinName : string;
      aDstModName : string;
      aDstPinName : string;
      T           : string;
    begin
      if Assigned( Peer)
      then begin
        aSrcModName := Peer.ModuleName   ( SrcModule);
        aSrcPinName := Peer.OutputPinName( SrcModule, SrcPin);
        aDstModName := Peer.ModuleName   ( DstModule);
        aDstPinName := Peer.InputPinName ( DstModule, DstPin);

        T := Format( '%sconnect( %s %s %s %s)' + sLineBreak, [ Indent( anIndent), aSrcModName, aSrcPinName, aDstModName, aDstPinName], AppLocale);
        S := S + T
      end;
    end;


    function    TConnection.IsSameConnection( aConnection: TConnection): Boolean;
    begin
      Result := False;

      if Assigned( aConnection)
      then
        Result :=
          ( SrcModule = aConnection.SrcModule) and
          ( SrcPin    = aConnection.SrcPin   ) and
          ( DstModule = aConnection.DstModule) and
          ( DstPin    = aConnection.DstPin   );
    end;


{ ========
  TConnections = class
  private
    FPeer       : TSynthPatch;
    FConnections: TConnectionArray;
  public
    property    Count: Integer                             read GetCount;
    property    Connection[ anIndex: Integer]: TConnection read GetConnection write SetConnection;
  private
}

    function    TConnections.GetCount: Integer;
    begin
      Result := Length( FConnections);
    end;


    function    TConnections.GetConnection( anIndex: Integer): TConnection;
    begin
      Result := FConnections[ anIndex]
    end;


    procedure   TConnections.SetConnection( anIndex: Integer; const aValue: TConnection);
    begin
      FConnections[ anIndex] := aValue;
    end;


//  public

    constructor TConnections.Create( aPeer: TSynthPatch);
    begin
      inherited Create;
      FPeer := aPeer;
    end;


    destructor  TConnections.Destroy; // override;
    begin
      Clear;
      inherited;
    end;


    procedure   TConnections.Save( anIndent: Integer; const aName: string; var S: string);
    var
      i : Integer;
    begin
      S := S + Indent( anIndent) + aName + '(' + sLineBreak;

      for i := 0 to Count - 1
      do Connection[ i].Save( anIndent + 1, S);

      S := S + Indent( anIndent) + ')' + sLineBreak;
    end;


    procedure   TConnections.Clear;
    begin
      while Count > 0
      do RemoveConnection( Count - 1);
    end;


    procedure   TConnections.AddConnection( const aValue: TConnection); // overload;
    begin
      SetLength( FConnections, Count + 1);
      FConnections[ Count - 1] := aValue;
    end;


    procedure   TConnections.AddConnection( aSrcModule, aSrcPin, aDstModule, aDstPin: Integer);  // overload;
    var
      aConnection: TConnection;
    begin
      aConnection := TConnection.Create;
      aConnection.Connect( FPeer, Count, aSrcModule, aSrcPin, aDstModule, aDstPin);
      AddConnection( aConnection);
    end;


    procedure   TConnections.RemoveConnection( anIndex: Integer); // overload;
    var
      aConnection: TConnection;
      i          : Integer;
    begin
      aConnection := Connection[ anIndex];

      if Assigned( aConnection)
      then begin
        for i := anIndex to Count - 2
        do Connection[ i] := Connection[ i + 1];

        SetLength( FConnections, Count - 1);
        aConnection.DisposeOf;
      end;
    end;


    function    TConnections.IsSameConnections( const aConnections: TConnections): Boolean;
    var
      i : Integer;
    begin
      Result := False;

      if Assigned( aConnections) and ( Count = aConnections.Count)
      then begin
        Result := True;

        for i := 0 to Count - 1
        do begin
          if not Connection[ i].IsSameConnection( aConnections.Connection[ i])
          then begin
            Result := False;
            Break;
          end;
        end;
      end;
    end;


{ ========
  TLinks = record
  private
    FSource : PSignal;
    FLinks  : array of PSignal;
  public
    property    Source : PSignal read FSource;
    property    Count  : Integer read GetCount;
  public
}

    procedure   TLinks.Initialize( aSource: PSignal);
    begin
      FSource := aSource;
      FLinks  := nil;
      FCount  := 0;
    end;


    procedure   TLinks.Uninitialize;
    begin
      if Assigned( FLinks)
      then FreeMem( FLinks);

      FLinks := nil;
      FCount := 0;
    end;


    procedure   TLinks.AddLink( aDestination: PSignal);
    var
      anOldLinks: PSignals;
    begin
      anOldLinks := FLinks;
      GetMem( FLinks, ( FCount + 1) * SizeOf( PSignal));

      if Assigned( anOldLinks)
      then begin
        Move( anOldLinks^, FLinks^, FCount * SizeOf( PSignal));
        FreeMem( anOldLinks);
      end;

      FLinks^[ FCount] := aDestination;
      Inc( FCount);
    end;


{ ========
  TOutputLinks = class
  private
    FOutLinks : array of TLinks;
  public
    property    Count                      : Integer read GetCount;
    property    OutLinks[ anIndex: Integer]: TLinks  read GetOutLinks;
  private
}


    function    TOutputLinks.FindSource( aSource: PSignal): Integer;
    var
      i : Integer;
    begin
      Result := -1;

      for i := 0 to FCount - 1
      do begin
        if FOutLinks^[ i].FSource = aSource
        then begin
          Result := i;
          Break;
        end;
      end;
    end;


    procedure   TOutputLinks.AddLink( aSource, aDestination: PSignal);
    var
      anIndex    : Integer;
      anOldLinks : PLinksArray;
    begin
      anIndex := FindSource( aSource);

      if anIndex < 0
      then begin
        anOldLinks := FOutLinks;
        GetMem( FOutLinks, ( FCount + 1) * SizeOf( TLinks));

        if Assigned( anOldLinks)
        then begin
          Move( anOldLinks^, FOutLinks^, FCount * SizeOf( TLinks));
          FreeMem( anOldLinks);
        end;

        FOutLinks^[ FCount].Initialize( aSource);
        anIndex := FCount;
        Inc( FCount);
      end;

      FOutLinks^[ anIndex].AddLink( aDestination);
    end;


//  public

    destructor  TOutputLinks.Destroy; // override;
    var
      i : Integer;
    begin
      for i := 0 to FCount - 1
      do FOutLinks^[ i].Uninitialize;

      FreeMem( FOutLinks);
      FOutLinks := nil;
      FCount    := 0;
    end;

    procedure   TOutputLinks.Compile( const aModule: TMod; const aPatch: TSynthPatch);
    var
      aModuleIndex : Integer;
      aConnections : TConnections;
      aConnection  : TConnection;
      i            : Integer;
      aDstModule   : TMod;
      aSource      : PSignal;
      aDestination : PSignal;
    begin
      if Assigned( FOutLinks)
      then begin
        for i := 0 to FCount - 1
        do FOutLinks^[ i].Uninitialize;

        FreeMem( FOutLinks);
        FOutLinks := nil;
        FCount    := 0;
      end;

      if Assigned( aModule) and Assigned( aPatch)
      then begin
        aConnections := aPatch.FConnections;

        if Assigned( aConnections)
        then begin
          aModuleIndex := aPatch.FindModule( aModule.Name);

          if aModuleIndex >= 0
          then begin
            for i := 0 to aConnections.Count - 1
            do begin
              aConnection := aConnections.Connection[ i];

              if aConnection.SrcModule = aModuleIndex
              then begin
                aDstModule   := aPatch.Module[ aConnection.DstModule];
                aSource      := @ aModule   .FOutputs[ aConnection.SrcPin];
                aDestination := @ aDstModule.FInputs [ aConnection.DstPin];
                AddLink( aSource, aDestination);
              end;
            end;
          end;
        end;
      end;
    end;


(* ========
  TMod = class
  private
    FParent          : TSynthPatch;       // The patch this module is part off
    FLocked          : Boolean;           // When true module should not be executed
    FExecuting       : Boolean;           // True when the module is currently being executed
    FName            : string;            // A unique name for this module - unique within it's parent
    FResetFlag       : Integer;           // Set when a module reset was requested
    FSamplesFast     : Int64;             // Sample counter, incremented each time a new Fast sample is calcultaed
    FSamplesSlow     : Int64;             // Sample counter, incremented each time a new Slow sample is calcultaed

  {$IFDEF PATCH_PROFILER}
  private
    FClocksFast      : Int64;             // CPU Fast clocks for current cycle
    FClocksSlow      : Int64;             // CPU Slow clocks for current cycle
    FTotalClocksFast : Int64;             // CPU Fast clocks accumulated over the module life time
    FTotalClocksSlow : Int64;             // CPU Slow clocks accumulated over the module life time
    FZips            : Int64;             // CPU clocks spent on dezoppering for the current cycle
    FTotalZips       : Int64;             // CPU clocks acumulated FZips over the module life time
  {$ENDIF}

  protected
    FIsFast          : Boolean;           // True when fast mode processing must be performed
    FIsSpedUp        : Boolean;           // True when module is normally slow but sped up due to fast signals being present on an input.
    FIsSlow          : Boolean;           // True when slow mode processing must be performed
    FIsMidi          : Boolean;           // True when MIDI reception must be enabled - for MIDI modules only
    FIsOSC           : Boolean;           // True when OSC  reception must be enabled - for OSC  modules only
    FInputCount      : Integer;           // Amount of input and amount of dezipper signals
    FOutputCount     : Integer;           // Amount of output signals
    FInputs          : TSignalArray;      // The input signals, these include both patch signals and user controls
    FDezips          : TSignalArray;      // An extra set of inputs for user controls, these are copied to FInCache with some filtering when the signal happens to be be in the deziper map
    FOutputs         : TSignalArray;      // The ouput signals.
    FDezipperMap     : TByteSet;          // Determines  what signals should be dezippered, most, but not all, knobs should be dezippered
  public
    property    Locked                                    : Boolean     read FLocked           write SetLocked;
    property    Name                                      : string      read FName             write SetName;
    property    ResetFlag                                 : Boolean     read GetResetFlag      write SetResetFlag;
    property    SamplesFast                               : Int64       read FSamplesFast;
    property    SamplesSlow                               : Int64       read FSamplesSlow;
    property    InputCount                                : Integer     read FInputCount;
    property    OutputCount                               : Integer     read FOutputCount;
    property    Output   [ anIndex: Integer]              : TSignal     read GetOutput;
    property    IsFast                                    : Boolean     read FIsFast;
    property    IsSlow                                    : Boolean     read FIsSlow;
    property    IsMidi                                    : Boolean     read FIsMidi;
    property    IsOSC                                     : Boolean     read FIsOSC;
    property    IsSpedUp                                  : Boolean     read FIsSpedUp;
  public
    property    InternalValue      [ const aName: string] : TSignal                            write SetInternalValue;
    property    AutomationValue    [ const aName: string] : TSignal                            write SetAutomationValue;
    property    InternalStringValue[ const aName: string] : string                             write SetInternalStringValue;
    property    StringValue        [ const aName: string] : string      read GetStringValue    write SetStringValue;

  {$IFDEF PATCH_PROFILER}
  public
    property    TotalClocksFast                           : Int64       read FTotalClocksFast;
    property    TotalClocksSlow                           : Int64       read FTotalClocksSlow;
    property    TotalZips                                 : Int64       read FTotalZips;
  {$ENDIF}

  private
*)

    function    TMod.AssertUniqueName( const aName: string): string; // virtual;
    begin
      if Assigned( FParent)
      then Result := FParent.AssertUniqueName( aName)
      else Result := aName;
    end;


//  private

    procedure   TMod.SetLocked( aValue: Boolean);
    // Note: Do not use a monitor construct here .. it is too slow ...
    begin
      if aValue
      then begin
        FLocked := True;                         // Just claim the lock ...

        while FExecuting                         // ... but then wait for any execution to finish, which will be quick,
        do ;                                     // and as long as Locked is True execution will not be re-entered.
      end
      else FLocked := False;                     // The lock can be released at all times
    end;


    procedure   TMod.SetName( const aValue: string);
    begin
      if aValue <> FName
      then FName := AssertUniqueName( aValue);
    end;


    function    TMod.GetResetFlag: Boolean; // virtual;
    begin
      Result := FResetFlag > 0;
    end;


    procedure   TMod.SetResetFlag( aValue: Boolean); // virtual;
    begin
      if aValue
      then FResetFlag := 1
      else begin
        if FResetFlag > 0
        then Dec( FResetFlag);
      end;
    end;


    function    TMod.GetOutput( anIndex: Integer): TSignal;
    begin
      if ( anIndex >= 0) and ( anIndex < OutputCount)
      then Result := FOutputs[ anIndex]
      else Result := 0.0;
    end;


    procedure   TMod.SetCommonValue( anIndex: Integer; aValue: TSignal; Dezipper: Boolean); // virtual;
    begin
      if anIndex >= 0
      then begin
        if Dezipper and ( anIndex in FDezipperMap)
        then FDezips[ anIndex] := aValue
        else begin
          if FInputs[ anIndex] = aValue
          then Pulse( anIndex)
          else FInputs[ anIndex] := aValue;
        end;
      end;
    end;


    procedure   TMod.SetInternalValue( const aName: string; aValue: TSignal); // virtual;
    var
      anIndex: Integer;
    begin
      anIndex := FindInput( aName);

      if anIndex >= 0
      then SetCommonValue( anIndex, aValue, DO_DEZIPPER)
      else SetInternal( aName, aValue);
    end;


    procedure   TMod.SetInternalStringValue( const aName, aValue: string);
    var
      aSignal : TSignal;
    begin
      if ParseTextValue( aValue, aSignal)
      then InternalValue[ aName] := aSignal
      else raise EUnitsUndefined.CreateFmt( 'Unparsable value ''%s''', [ aName])
    end;


    procedure   TMod.SetAutomationValue( const aName: string; aValue: TSignal); // virtual;
    var
      anIndex: Integer;
    begin
      anIndex := FindInput( aName);

      if anIndex >= 0
      then SetCommonValue( anIndex, aValue, NO_DEZIPPER)
      else SetInternal( aName, aValue);
    end;


//  protected

    procedure   TMod.SetStringValue( const aName, aValue: string); // virtual;
    begin
      // Nothing here.
    end;


    function    TMod.GetStringValue( const aName: string): string; // virtual;
    begin
      Result := '';
    end;


//  protected

    procedure   TMod.FixInputs;
    var
      aCount : Integer;
    begin
      aCount := 0;

      while not SameText( InputName( aCount), NO_INPUT)
      do begin
        Inc( aCount);
        if aCount > MAX_INPUTS
        then raise EInputNotFound.CreateFmt( 'are there really over %d inputs? (probably %s.InputName is broken and never returns NO_INPUT)', [ MAX_INPUTS, ClassName]);
      end;

      SetLength( FInputs    , aCount);
      SetLength( FDezips    , aCount);
      FInputCount := aCount;
    end;


    procedure   TMod.FixOutputs;
    var
      aCount : Integer;
    begin
      aCount := 0;

      while not SameText( OutputName( aCount), NO_OUTPUT)
      do begin
        Inc( aCount);
        if aCount > MAX_OUTPUTS
        then raise EOutputNotFound.CreateFmt( 'are there really over %d outputs? (probably %s.OutputName is broken and never returns NO_OUTPUT)', [ MAX_OUTPUTS, ClassName]);
      end;

      SetLength( FOutputs, aCount);
      FOutputCount := aCount;
    end;


    procedure   TMod.HandleParam( const aControl: string; aValue: TSignal); // virtual;
    // Handle a parameter set from the editor patch
    begin
      InternalValue[ aControl] := aValue;
    end;


    procedure   TMod.HandleAutomationParam( const aControl: string; aValue: TSignal); // virtual;
    begin
      AutomationValue[ aControl] := aValue;
    end;


    procedure   TMod.Log( const aMsg: string); // virtual;
    begin
      if Assigned( FParent)
      then FParent.Log( aMsg);
    end;


    procedure   TMod.LogFmt( const aFmt: string; const anArgs: array of const);
    begin
      Log( Format( aFmt, anArgs, AppLocale));
    end;


    procedure   TMod.Error( const aMsg: string); // virtual;
    begin
      if Assigned( FParent)
      then FParent.Error( aMsg);
    end;


    procedure   TMod.ErrorFmt( const aFmt: string; const anArgs: array of const);
    begin
      Error( Format( aFmt, anArgs, AppLocale));
    end;


    {$IFDEF PATCH_PROFILER}

//  protected

    procedure   TMod.PreTickFast;
    begin
      {$Q-R-}
      FClocksFast := GetCpuClockCycleCount;
      {$Q+R+}
    end;


    procedure   TMod.PostTickFast;
    begin
      {$Q-R-}
      FTotalClocksFast := FTotalClocksFast + GetCpuClockCycleCount - FClocksFast;
      {$Q+R+}
    end;


    procedure   TMod.PreTickSlow;
    begin
      {$Q-R-}
      FClocksSlow := GetCpuClockCycleCount;
      {$Q+R+}
    end;


    procedure   TMod.PostTickSlow;
    begin
      {$Q-R-}
      FTotalClocksSlow := FTotalClocksSlow + GetCpuClockCycleCount - FClocksSlow;
      {$Q+R+}
    end;


    procedure   TMod.PreZipper;
    begin
      {$Q-R-}
      FZips := GetCpuClockCycleCount;
      {$Q+R+}
    end;


    procedure   TMod.PostZipper;
    begin
      {$Q-R-}
      FTotalZips := FTotalZips + GetCpuClockCycleCount - FZips;
      {$Q+R+}
    end;


    procedure   TMod.ClearProfile;
    begin
      FTotalClocksFast := 0;
      FTotalClocksSlow := 0;
      FSamplesFast     := 0;
      FSamplesSlow     := 0;
      FTotalZips       := 0;
    end;

    {$ENDIF}


//  public

    constructor TMod.Create( aParent: TSynthPatch; const aName: string); // virtual;
    begin
      inherited Create;
      {$IFDEF PATCH_PROFILER}
      FTotalClocksSlow := 0;
      FTotalClocksFast := 0;
      FTotalZips       := 0;
      {$ENDIF}
      FIsFast          := False;
      FIsSlow          := False;
      FIsMidi          := False;
      FIsOSC     := False;
      FParent          := aParent;
      Name             := aName;
      FOutputLinks     := TOutputLinks.Create;
      FixOutputs;
      FixInputs;
      SetDefaults;
    end;


    destructor  TMod.Destroy; // override;
    begin
      FParent := nil;
      FreeAndNil( FOutputLinks);
      inherited;
    end;


    procedure   TMod.CompileLinks; // virtual;
    begin
      Locked := true;
      try
        FOutputLinks.Compile( Self, FParent);
      finally
        Locked := False;
      end;
    end;


    procedure   TMod.PropagateSignals;
    var
      i : Integer;
      j : Integer;
    begin
      for i := 0 to FOutputLinks.FCount - 1
      do begin
        for j := 0 to FOutputLinks.FOutLinks^[ i].FCount - 1
        do FOutputLinks.FOutLinks^[ i].FLinks^[ j]^ := FOutputLinks.FOutLinks^[ i].FSource^;
      end;
    end;


    procedure   TMod.SetDefaults; // virtual;
    begin
    end;


    procedure   TMod.SaveKeyValue( anIndent: Integer; const aName: string; aValue: TSignal; var S: string);
    var
      aBaseName : string;
      i         : Integer;
      Output    : string;
    begin
      aBaseName := '';

      for i := 1 to Length( aName)
      do begin
        if CharInSet( aName[ i], IdentifierStarters)
        then aBaseName := abaseName + aName[ i]
        else Break;
      end;

      Output := Format( '%s', [ SignalToStr( aValue)], AppLocale);
      S := S + Format( Indent( anIndent + 1) + '%s %s', [ aName, Output], AppLocale);
    end;


    procedure   TMod.SaveInputValues( anIndent: Integer; var S: string);
    var
      i : Integer;
      N : string;
      p : Integer;
    begin
      p := 1;
      S := S + Indent( anIndent) + 'inputs(' + sLineBreak;

      for i := 0 to InputCount - 1
      do begin
        N := InputName( i);

        if SameText( N, NO_INPUT)
        then begin
          N := Format( 'unnamed_input_%d', [ p], AppLocale);
          Inc( p);
        end;

        SaveKeyValue( anIndent, N, FInputs[ i], S);
        S := S + sLineBreak;
      end;

      S := S + Indent( anIndent) + ')' + sLineBreak;
    end;


    procedure   TMod.SaveOutputValues( anIndent: Integer; var S: string);
    var
      i : Integer;
      N : string;
      p : Integer;
    begin
      p := 1;
      S := S + Indent( anIndent) + 'outputs(' + sLineBreak;

      for i := 0 to OutputCount - 1
      do begin
        N := OutputName( i);

        if SameText( N, NO_OUTPUT)
        then begin
          N := Format( 'unnamed_output_%d', [ p], AppLocale);
          Inc( p);
        end;

        SaveKeyValue( anIndent, N, FOutputs[ i], S);
        S := S + sLineBreak;
      end;

      S := S + Indent( anIndent) + ')' + sLineBreak;
    end;


    procedure   TMod.SaveDetails( anIndent: Integer; var S: string); // virtual;
    begin
      S := S + Format ( Indent( anIndent) + '%s' + sLineBreak, [ ClassName ], AppLocale);
      S := S + Format ( Indent( anIndent) + '%s' + sLineBreak, [ Name      ], AppLocale);
      SaveInputValues ( anIndent, S);
      SaveOutputValues( anIndent, S);
    end;


    procedure   TMod.SetInternal( const aName: string; aValue: TSignal); // virtual;
    begin
      // Internal is an option not present here.
    end;


    function    TMod.GetInternal( const aName: string): TSignal; // virtual;
    begin
      // Internals is an option not present here.
      Result := UNDEFINED;
    end;


    procedure   TMod.Save( anIndent: Integer; var S: string);
    begin
      S := S + Indent( anIndent) + 'module(' + sLineBreak;
      SaveDetails( anIndent + 1, S);
      S := S + Indent( anIndent) + ')' + sLineBreak;
    end;


    procedure   TMod.Reset; // virtual;
    begin
      ResetFlag := True;
    end;


    procedure   TMod.ResetSampleCounts;
    begin
      FSamplesFast := 0;
      FSamplesSlow := 0;
    end;


    procedure   TMod.Clear; // virtual;
    begin
    end;


    procedure   TMod.Panic; // virtual;
    begin
      // No panic here ...
    end;


    procedure   TMod.Pulse( anInput: Integer); // virtual;
    begin
      // No pulsing here ...
    end;


    function    TMod.FindInput( const aPinName: string): Integer; // virtual;
    var
      i : Integer;
    begin
      Result := UNDEF_PMC;

      for i := 0 to InputCount - 1
      do begin
        if SameText( aPinName, InputName( i))
        then begin
          Result := i;
          Break;
        end;
      end;
    end;


    function    TMod.FindOutput( const aPinName: string): Integer; // virtual;
    var
      i : Integer;
    begin
      Result := UNDEF_PMC;

      for i := 0 to OutputCount - 1
      do begin
        if SameText( aPinName, OutputName( i))
        then begin
          Result := i;
          Break;
        end;
      end;
    end;


    function    TMod.InputName( anIndex: Integer): string; // virtual;
    begin
      Result := NO_INPUT;
    end;


    function    TMod.OutputName( anIndex: Integer): string; // virtual;
    begin
      Result := NO_OUTPUT;
    end;


    procedure   TMod.MakeFast;
    begin
      FIsSpedUp := True;
    end;


    procedure   TMod.Tick;
    begin
      {$Q-R-}
      Inc( FSamplesFast);
      {$Q+R+}
      {$IFDEF PATCH_PROFILER}PreTickFast;{$ENDIF}
      DoTick;
      {$IFDEF PATCH_PROFILER}PostTickFast;{$ENDIF}
    end;


    procedure   TMod.SlowTick;
    begin
      {$Q-R-}
      Inc( FSamplesSlow);
      {$Q+R+}
      {$IFDEF PATCH_PROFILER}PreTickSlow;{$ENDIF}
      DoSlowTick;
      {$IFDEF PATCH_PROFILER}PostTickSlow;{$ENDIF}
    end;


    procedure   TMod.DeZipper;
    const
      Speed = 1.0 / 512.0;
    var
      i : Integer;
    begin
      {$IFDEF PATCH_PROFILER}PreZipper;{$ENDIF}
      for i := 0 to Length( FInputs) - 1
      do begin
        if i in FDezipperMap
        then begin
          FInPuts[ i] := Normalize( Speed * ( FDezips[ i] - FInputs[ i]) + FInputs[ i]);

          if Abs( FInputs[ i]) < MUTE_LIMIT
          then FInputs[ i] := 0;
        end;
      end;
      {$IFDEF PATCH_PROFILER}PostZipper;{$ENDIF}
    end;


    procedure   TMod.DoTick; // virtual;
    begin
      // audio rate
    end;


    procedure   TMod.DoSlowTick; // virtual;
    begin
      // control rate
    end;


    procedure   TMod.FixZippers; // virtual;
    var
      i : Integer;
    begin
      for i := 0 to Length( FInputs) - 1
      do begin
        if i in FDezipperMap
        then FInputs[ i] := FDezips[ i];
      end;
    end;


    procedure   TMod.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // virtual;
    begin
      // Collect info to be sent to module lights - this module has no lights
    end;


    procedure   TMod.GatherData( const aPrefix: string; const aCallback: TDataHandler); // virtual;
    begin
      // Collect numerical info to be sent to module viewers - this module has no viewers
    end;


    procedure   TMod.GatherXYData( const aPrefix: string; const aCallback: TXYDataHandler); // virtual;
    begin
      // Collect numerical info to be sent to module XY viewers - this module has no XY viewers
    end;


    procedure   TMod.GatherInfo( const aPrefix: string; const aCallback: TInfoHandler); // virtual;
    begin
      // Collect string info to be sent to module viewers - this module has no viewers
    end;


    procedure   TMod.AcceptParam( const aPath: string; aValue: TSignal);
    // Accept knob data sent from visual modules
    var
      aPrefix : string;
      aParam  : string;
    begin
      if ( ClassType = TSynthPatch)
      then HandleParam( aPath, aValue)
      else begin
        ParsePrefix( aPath, aPrefix, aParam);

        if SameText( aPrefix, Name)                      // is the message for us?
        then begin
          if Pos( '/', aParam) = 0                       // When no more slashes left
          then begin                                     // we must be at the final bit ...
            aParam := ParsePostfix( aParam);
            HandleParam( aParam, aValue);
          end;
        end;
      end;
    end;


    procedure   TMod.AcceptAutomationParam( const aPath: string; aValue: TSignal);
    // Accept knob data sent from visual modules
    var
      aPrefix : string;
      aParam  : string;
    begin
      if ( ClassType = TSynthPatch)
      then HandleAutomationParam( aPath, aValue)
      else begin

        ParsePrefix( aPath, aPrefix, aParam);

        if SameText( aPrefix, Name)                      // is the message for us?
        then begin
          if Pos( '/', aParam) = 0                       // When no more slashes left
          then begin                                     // we must be at the final bit ...
            aParam := ParsePostfix( aParam);
            HandleAutomationParam( aParam, aValue);
          end;
        end;
      end;
    end;


    function    TMod.GetProfileInfoHtml: string;
    {$IFDEF PATCH_PROFILER}
    var
      FR : string;
      SR : string;
    {$ENDIF}
    begin
    {$IFDEF PATCH_PROFILER}
      if SamplesFast = 0
      then FR := '---'
      else FR := Format( '%.2f', [ TotalClocksFast / SamplesFast], AppLocale);
      if SamplesSlow = 0
      then SR := '---'
      else SR := Format( '%.2f', [ TotalClocksSlow / SamplesSlow / CONTROL_DEC], AppLocale);

      Result :=
        Format(
          '<tr><td>%s</td> <td>%s</td> <td align="right">%d</td> <td align="right">%d</td> <td align="right">%d</td> <td align="right">%d</td> <td align="right">%s</td> <td align="right">%s</td></tr>',
          [
            Name,
            ClassName,
            SamplesFast,
            SamplesSlow,
            TotalClocksFast,
            TotalClocksSlow,
            FR,
            SR
          ],
          AppLocale
        );
      {$ELSE}
        Result := '';
      {$ENDIF}
    end;


    function    TMod.IsSameModule( const aModule: TMod): Boolean;
    begin
      Result := False;

      if Assigned( aModule)
      then Result := ClassType = aModule.ClassType;
    end;


    procedure   TMod.AcceptMidi( const aMsg: TMidiMessage); // virtual;
    begin
      // Nothing here ... for inheritance only
    end;


    procedure   TMod.SendMidi( const aMsg: TMidiMessage);
    begin
      if Assigned( FParent)
      then FParent.SendMidi( aMsg);
    end;


    function    TMod.OSCMatched( const aMsg: TOSCMessage): Boolean; // virtual;
    begin
      Result := False;
    end;


    procedure   TMod.AcceptOSC( const aSynthName: string; const aMsg: TOSCPacket); // virtual;
    begin
      // Nothing here, if is for children, just building daydreams ...
    end;


    procedure   TMod.SendOSCBytes( const aMsg: TBytes); // virtual;
    begin
      if Assigned( FParent)
      then FParent.SendOSCBytes( aMsg);
    end;


    procedure   TMod.SampleRateChanged; // virtual;
    // Notification for sample rate change, the actual values are in global variables
    // System_Rate for the audio rate and Control_Rate for the control rate.
    // Modules caching sample rate dependent data should recalculate such data in an
    // overriden version of this procedure.
    begin
      // Nothing for abstract modules.
    end;


    procedure   TMod.CollectDenormals( const aRange: TSignal; var aMap: TPinMap);

      function RegardAsDenormal( const aValue: TSignal): Boolean;
      type
        TWordArray = array[ 0 .. 3] of Word;
        PWordArray = ^TWordArray;
      begin
        Result := ((( PWordArray( @ aValue)^[ 3] and $7ff0) = 0) or ( Abs( aValue) < aRange)) and ( aValue <> 0);
      end;

    var
      i : Integer;
    begin
      aMap.Ins  := [];
      aMap.Outs := [];

      for i := 0 to InputCount - 1
      do begin
        if RegardAsDenormal( FInputs[ i])
        then aMap.Ins := aMap.Ins + [ i];
      end;

      for i := 0 to OutputCount - 1
      do begin
        if RegardAsDenormal( FOutputs[ i])
        then aMap.Outs := aMap.Outs + [ i];
      end;
    end;


{ ========
  TModInput = class( TMod)
  public
}

    function    TModInput.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        o_out1 : Result := 'out1';
        o_out2 : Result := 'out2';
        o_out3 : Result := 'out3';
        o_out4 : Result := 'out4';
        o_out5 : Result := 'out5';
        o_out6 : Result := 'out6';
        o_out7 : Result := 'out7';
        o_out8 : Result := 'out8';
        else     Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModOutput = class( TMod)
  public
}

    function    TModOutput.InputName ( anIndex: Integer): string; // override;
    begin
      case anIndex of
        i_in1 : Result := 'in1';
        i_in2 : Result := 'in2';
        i_in3 : Result := 'in3';
        i_in4 : Result := 'in4';
        i_in5 : Result := 'in5';
        i_in6 : Result := 'in6';
        i_in7 : Result := 'in7';
        i_in8 : Result := 'in8';
        else    Result := NO_INPUT;
      end;
    end;


(* ========
  TSynthPatch = class
  private
    FVersion         : Integer;
    FMidiChannel     : Byte;
    FModules         : TModuleArray;
    FConnections     : TConnections;
    FDecimatorCount  : Integer;
    FInputFifo       : TSignalOctetFifo;
    FOnSendMidi      : TOnMidiMessage;
    FOnSendOscString : TOnOSCString;
    FOnSynthLog      : TOnSynthLog;
    FOnSynthError    : TOnSynthError;
  public
    property    Version                   : Integer        read FVersion           write FVersion;
    property    ModuleCount               : Integer        read GetModuleCount;
    property    Module[ anIndex: Integer] : TMod           read GetModule;
    property    ConnectionCount           : Integer        read GetConnectionCount;
    property    MidiChannel               : Byte           read FMidiChannel       write FMidiChannel;
    property    Overflows                 : Integer        read GetOverflows;
    property    Underflows                : Integer        read GetUnderflows;
    property    OnSendMidi                : TOnMidiMessage read FOnSendMidi        write FOnSendMidi;
    property    OnSendOSCString           : TOnOSCString   read FOnSendOscString   write FOnSendOscString;
    property    OnSynthLog                : TOnSynthLog    read FOnSynthLog        write FOnSynthLog;
    property    OnSynthError              : TOnSynthError  read FOnSynthError      write FOnSynthError;
  private
*)

    function    TSynthPatch.AssertUniqueName( const aName: string): string; // override;
    var
      aModule : TMod;
    begin
      // calling inherited here would go recursive
      Result := aName;

      for aModule in FModules
      do begin
        if SameText( aModule.Name, aName)
        then raise EModule.CreateFmt( 'module name "%s" is not unique.', [ aName]);
      end;
    end;


//  private

    function    TSynthPatch.CheckModuleIndex( anIndex: Integer): Boolean;
    begin
      Result := ( anIndex >= 0) and ( anIndex < ModuleCount);
    end;


    function    TSynthPatch.GetModuleCount: Integer;
    begin
      Result := Length( FModules);
    end;


    function    TSynthPatch.GetModule( anIndex: Integer): TMod;
    begin
      Result := nil;

      if CheckModuleIndex( anIndex)
      then Result := FModules[ anIndex];
    end;


    function    TSynthPatch.GetConnectionCount: Integer;
    begin
      Result := 0;

      if Assigned( FConnections)
      then Result := FConnections.Count;
    end;


    procedure   TSynthPatch.SetResetFlag( aValue: Boolean); // override;
    begin
      if aValue
      then FResetFlag := 2 * CONTROL_DEC + 1
      else begin
        if FResetFlag > 0
        then Dec( FResetFlag);
      end;
    end;


    function    TSynthPatch.GetOverflows: Integer;
    begin
      Result := 0;

      if Assigned( FInputFifo)
      then Result := Result + FInputFifo.Overflows;
    end;


    function    TSynthPatch.GetUnderflows: Integer;
    begin
      Result := 0;

      if Assigned( FInputFifo)
      then Result := Result + FInputFifo.Underflows;
    end;


    function    TSynthPatch.FixPrefix( const aPrefix: string): string;
    begin
      if aPrefix = ''
      then Result := Name
      else Result := aPrefix + '/' + Name;
    end;


//  protected

    procedure   TSynthPatch.HandleParam( const aPath: string; aValue: TSignal); // override;
    var
      aModule : TMod;
    begin
      for aModule in FModules
      do aModule.AcceptParam( aPath, aValue);
    end;


    procedure   TSynthPatch.HandleAutomationParam( const aPath: string; aValue: TSignal); // override;
    var
      aModule : TMod;
    begin
      for aModule in FModules
      do aModule.AcceptAutomationParam( aPath, aValue);
    end;


    procedure   TSynthPatch.Log( const aMsg: string); // override;
    begin
      if Assigned( FOnSynthLog)
      then FOnSynthLog( Self, aMsg);
    end;


    procedure   TSynthPatch.Error( const aMsg: string); // override;
    begin
      if Assigned( FOnSynthError)
      then FOnSynthError( Self, aMsg);
    end;


//  public

    constructor TSynthPatch.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast      := True;
      FIsSlow      := False;
      FIsMidi      := True;
      FIsOSC       := True;
      FVersion     := 0;
      FInputFifo   := TSignalOctetFifo.Create( FIFO_SIZE);
      FConnections := TConnections.Create( Self);
    end;


    destructor  TSynthPatch.Destroy; // override;
    begin
      FreeAndNil( FConnections);  // Do this first for a faster clear
      Clear;                      // Clears the modules - that is it removes and frees them
      FreeAndNil( FInputFifo);
      inherited;
    end;


    procedure   TSynthPatch.CompileLinks; // override;
    var
      aModule : TMod;
    begin
      for aModule in FModules
      do aModule.CompileLinks;
    end;


    procedure   TSynthPatch.SetDefaults; // override;
    var
      aModule : TMod;
    begin
      for aModule in FModules
      do aModule.SetDefaults;
    end;


    procedure   TSynthPatch.SaveConnections( anIndent: Integer; var S: string);
    begin
      if Assigned( FConnections)
      then FConnections.Save( anIndent, 'connections', S);
    end;


    procedure   TSynthPatch.SaveModules( anIndent: Integer; var S: string);
    var
      aModule : TMod;
    begin
      S := S + Indent( anIndent) + 'modules(' + sLineBreak;

      for aModule in FModules
      do aModule.Save( anIndent + 1, S);

      S := S + Indent( anIndent) + ')' + sLineBreak;
    end;


    procedure   TSynthPatch.SaveDetails( anIndent: Integer; var S: string); // override;
    begin
      inherited;
      S := S + Indent( anIndent) + 'version ' + IntToStr( Version) + sLineBreak;
      SaveModules    ( anIndent, S);
      SaveConnections( anIndent, S);
    end;


    procedure   TSynthPatch.Clear;   // Clear moduleas and connections
    begin
      ClearModules;
      inherited Clear;
    end;


    procedure   TSynthPatch.ResetSampleCounts; // override;
    var
      aModule : TMod;
    begin
      for aModule in FModules
      do aModule.ResetSampleCounts;

      inherited;
    end;


    procedure   TSynthPatch.ClearModules;
    var
      aModule : TMod;
    begin
      while ModuleCount > 0
      do begin
        aModule :=  FModules[ ModuleCount - 1];
        SetLength(  FModules, ModuleCount - 1);
        aModule.DisposeOf;
      end;
    end;


    procedure   TSynthPatch.Panic; // override;
    var
      aModule : TMod;
    begin
      for aModule in FModules
      do aModule.Panic;
    end;


    function    TSynthPatch.FindModule( const aModuleName: string): Integer;
    var
      i : Integer;
    begin
      Result := UNDEF_PMC;

      for i := 0 to ModuleCount - 1
      do begin
        if SameText( aModuleName, Module[ i].Name)
        then begin
          Result := i;
          Break;
        end;
      end;
    end;


    function    TSynthPatch.ModuleName( anIndex: Integer): string;
    begin
      if ( anIndex >= 0) and ( anIndex < ModuleCount)
      then Result := Module[ anIndex].Name
      else Result := NO_MODULE;
    end;


    function    TSynthPatch.FindInputPin( const aModuleIndex: Integer; const aPinName: string): Integer;
    begin
      Result := UNDEF_PMC;

      if aModuleIndex >= 0
      then Result := Module[ aModuleIndex].FindInput( aPinName);
    end;


    function    TSynthPatch.FindOutputPin( const aModuleIndex: Integer; const aPinName: string): Integer;
    begin
      Result := UNDEF_PMC;

      if aModuleIndex >= 0
      then Result := Module[ aModuleIndex].FindOutput( aPinName);
    end;


    function    TSynthPatch.InputPinName( aModuleIndex, aPinIndex: Integer): string;
    begin
      Result := NO_INPUT;

      if ( aModuleIndex >= 0) and ( aModuleIndex < ModuleCount)
      then Result := Module[ aModuleIndex].InputName( aPinIndex);
    end;


    function    TSynthPatch.OutputPinName( aModuleIndex, aPinIndex: Integer): string;
    begin
      Result := NO_OUTPUT;

      if ( aModuleIndex >= 0) and ( aModuleIndex < ModuleCount)
      then Result := Module[ aModuleIndex].OutputName( aPinIndex);
    end;


    procedure   TSynthPatch.AddModule( const aModule: TMod); // overload;
    begin
      SetLength( FModules, ModuleCount + 1);
      FModules[ ModuleCount - 1] := aModule;
    end;


    procedure   TSynthPatch.AddConnection( aSrcModule, aSrcPin, aDstModule, aDstPin: Integer); // overload;
    begin
      if Assigned( FConnections)
      then FConnections.AddConnection( aSrcModule, aSrcPin, aDstModule, aDstPin);
    end;


    procedure   TSynthPatch.AddConnection( const aSrcModule, aSrcPin, aDstModule, aDstPin: string); // overload;
    var
      aSrcM: Integer;
      aSrcP: Integer;
      aDstM: Integer;
      aDstP: Integer;
    begin
      aSrcM := FindModule( aSrcModule);

      if aSrcM >= 0
      then begin
        aSrcP := FindOutputPin( aSrcM, aSrcPin);

        if aSrcP >= 0
        then begin
          aDstM := FindModule( aDstModule);

          if aDstM >= 0
          then begin
            aDstP := FindInputPin( aDstM, aDstPin);

            if aDstP >= 0
            then AddConnection( aSrcM, aSrcP, aDstM, aDstP)
            else raise EInputNotFound.CreateFmt( 'input "%s" not found on destination module "%s"', [ aDstPin, aDstModule]);
          end
          else raise EModuleNotFound.CreateFmt( 'destination module "%s" not found', [ aDstModule]);
        end
        else raise EOutputNotFound.CreateFmt( 'output "%s" not found on source module "%s"', [ aSrcPin, aSrcModule]);
      end
      else raise EModuleNotFound.CreateFmt( 'source module "%s" not found', [ aSrcModule]);
    end;


const

  synthpatch_in1  = 0;
  synthpatch_in2  = 1;
  synthpatch_in3  = 2;
  synthpatch_in4  = 3;
  synthpatch_in5  = 4;
  synthpatch_in6  = 5;
  synthpatch_in7  = 6;
  synthpatch_in8  = 7;

  synthpatch_out1 = 0;
  synthpatch_out2 = 1;
  synthpatch_out3 = 2;
  synthpatch_out4 = 3;
  synthpatch_out5 = 4;
  synthpatch_out6 = 5;
  synthpatch_out7 = 6;
  synthpatch_out8 = 7;

    function    TSynthPatch.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        synthpatch_in1 : Result := 'in1';
        synthpatch_in2 : Result := 'in2';
        synthpatch_in3 : Result := 'in3';
        synthpatch_in4 : Result := 'in4';
        synthpatch_in5 : Result := 'in5';
        synthpatch_in6 : Result := 'in6';
        synthpatch_in7 : Result := 'in7';
        synthpatch_in8 : Result := 'in8';
        else             Result := NO_INPUT;
      end;
    end;


    function    TSynthPatch.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        synthpatch_out1 : Result := 'out1';
        synthpatch_out2 : Result := 'out2';
        synthpatch_out3 : Result := 'out3';
        synthpatch_out4 : Result := 'out4';
        synthpatch_out5 : Result := 'out5';
        synthpatch_out6 : Result := 'out6';
        synthpatch_out7 : Result := 'out7';
        synthpatch_out8 : Result := 'out8';
        else              Result := NO_OUTPUT;
      end;
    end;


    procedure   TSynthPatch.DoTick; // Override;
    var
      i         : Integer;
      p         : Integer;
      aModCount : Integer;
      aMod      : TMod;
      anOctet   : TSignalOctet;
      AudioCh1  : TSignal;
      AudioCh2  : TSignal;
      AudioCh3  : TSignal;
      AudioCh4  : TSignal;
      AudioCh5  : TSignal;
      AudioCh6  : TSignal;
      AudioCh7  : TSignal;
      AudioCh8  : TSignal;
      OldEM     : TArithmeticExceptionMask;
    begin
      try
        // Get patch external inputs

        anOctet   := FInputFifo.GetData;
        aModCount := ModuleCount;
        p         := 0;
        AudioCh1 := 0; AudioCh2 := 0; AudioCh3 := 0; AudioCh4 := 0;
        AudioCh5 := 0; AudioCh6 := 0; AudioCh7 := 0; AudioCh8 := 0;
        FDecimatorCount := ( FDecimatorCount + 1) and ( CONTROL_DEC - 1);

        {$IFDEF DEBUG_DENORMS}
        OldEM := SetExceptionMask( GetExceptionMask - [ exDenormalized]); // For debugging trap denormalized exceptions in addition to the usual ones
        {$ELSE}
        OldEM := SetExceptionMask( exAllArithmeticExceptions);            // For non-debug code disable all floating point exceptions
        {$ENDIF}

        try
          for i := 0 to aModCount - 1                                     // Tick the modules in the patch
          do begin
            aMod := Module[ i];

            if ResetFlag
            then aMod.Reset;

            aMod.FExecuting := True;                                        // Start with flagging being in execution so locked operations have to wait

            try
              if not aMod.Locked                                            // Do not execute locked modules, this may get set even when FEcecuting was,
              then begin                                                    // True, but as long as Executing is True locked code can not be executed.

                if aMod is TModInput                                        // When module is an input module it needs the current in sample
                then begin
                  aMod.FOutputs[ 0] := anOctet.Ch1;
                  aMod.FOutputs[ 1] := anOctet.Ch2;
                  aMod.FOutputs[ 2] := anOctet.Ch3;
                  aMod.FOutputs[ 3] := anOctet.Ch4;
                  aMod.FOutputs[ 4] := anOctet.Ch5;
                  aMod.FOutputs[ 5] := anOctet.Ch6;
                  aMod.FOutputs[ 6] := anOctet.Ch7;
                  aMod.FOutputs[ 7] := anOctet.Ch8;
                  aMod.PropagateSignals;
                end
                else if aMod is TModOutput                                  // When module is an output module add its inputs to our outputs
                then begin
                  AudioCh1 := AudioCh1 + Normalize( aMod.FInputs[ 0]);
                  AudioCh2 := AudioCh2 + Normalize( aMod.FInputs[ 1]);
                  AudioCh3 := AudioCh3 + Normalize( aMod.FInputs[ 2]);
                  AudioCh4 := AudioCh4 + Normalize( aMod.FInputs[ 3]);
                  AudioCh5 := AudioCh5 + Normalize( aMod.FInputs[ 4]);
                  AudioCh6 := AudioCh6 + Normalize( aMod.FInputs[ 5]);
                  AudioCh7 := AudioCh7 + Normalize( aMod.FInputs[ 6]);
                  AudioCh8 := AudioCh8 + Normalize( aMod.FInputs[ 7]);
                end
                else if aMod.FIsFast
                then begin
                  aMod.Tick;                                                // Normal fast mode
                  aMod.PropagateSignals;
                end
                else if aMod.FIsSpedUp
                then begin
                  aMod.SlowTick;                                            // Sped up slow mode for RateSmart modules
                  aMod.PropagateSignals;
                end;

                if p = FDecimatorCount
                then begin
                  if aMod.FIsSlow and not aMod.FIsSpedUp                    // Exclude sped up RateSmart modules
                  then begin
                    aMod.SlowTick;
                    aMod.PropagateSignals;
                  end;

                  aMod.DeZipper;                                            // Dezipper at control rate
                end;
              end;
            finally
              aMod.FExecuting := False;                                     // Allow for locked opertions to take place again
            end;

            p := ( p + 1) and ( CONTROL_DEC - 1);
          end;

          ResetFlag := False;

          FOutputs[ 0] := AudioCh1;                                       // Finally fix our own outputs
          FOutputs[ 1] := AudioCh2;
          FOutputs[ 2] := AudioCh3;
          FOutputs[ 3] := AudioCh4;
          FOutputs[ 4] := AudioCh5;
          FOutputs[ 5] := AudioCh6;
          FOutputs[ 6] := AudioCh7;
          FOutputs[ 7] := AudioCh8;
        finally
          ClearFPUExceptions( False);
          SetExceptionMask( OldEM);
        end;
      except
        on E: Exception
        do KilledException( E);                                           // Always kill all exceptions here.
      end;
    end;


    procedure   TSynthPatch.FixZippers; // override;
    var
      i  : Integer;
    begin
      inherited;

      for i := 0 to ModuleCount - 1
      do Module[ i].FixZippers;
    end;


    procedure   TSynthPatch.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    var
      aModule : TMod;
      S       : string;
    begin
      try
        S := FixPrefix( aPrefix);

        for aModule in FModules
        do aModule.GatherLights( S, aCallback);
      except
        on E: Exception
        do KilledException( E);
      end;
    end;


    procedure   TSynthPatch.GatherData( const aPrefix: string; const aCallback: TDataHandler); // override;
    var
      aModule : TMod;
      S       : string;
    begin
      try
        S := FixPrefix( aPrefix);

        for aModule in FModules
        do aModule.GatherData( S, aCallback);
      except
        on E: Exception
        do KilledException( E);
      end;
    end;


    procedure   TSynthPatch.GatherXYData( const aPrefix: string; const aCallback: TXYDataHandler); // override;
    var
      aModule : TMod;
      S       : string;
    begin
      try
        S := FixPrefix( aPrefix);

        for aModule in FModules
        do aModule.GatherXYData( S, aCallback);
      except
        on E: Exception
        do KilledException( E);
      end;
    end;


    procedure   TSynthPatch.GatherInfo( const aPrefix: string; const aCallback: TInfoHandler); // override;
    var
      aModule : TMod;
      S       : string;
    begin
      try
        S := FixPrefix( aPrefix);

        for aModule in FModules
        do aModule.GatherInfo( S, aCallback);
      except
        on E: Exception
        do KilledException( E);
      end;
    end;


    procedure   TSynthPatch.AcceptSampleOctet( aCh1, aCh2, aCh3, aCh4, aCh5, aCh6, aCh7, aCh8: TSignal);
    begin
      FInputFifo.Append( aCh1, aCh2, aCh3, aCh4, aCh5, aCh6, aCh7, aCh8);
    end;


    function    TSynthPatch.FifoFillCount: Integer;
    begin
      Result := FInputFifo.FillCount;
    end;


    procedure   TSynthPatch.ClearProfile;
    {$IFDEF PATCH_PROFILER}
    var
      aModule : TMod;
    {$ENDIF}
    begin
    {$IFDEF PATCH_PROFILER}
      for aModule in FModules
      do aModule.ClearProfile;
    {$ENDIF}
    end;


    function    TSynthPatch.CreateProfileHtml: TStringList;
    {$IFDEF PATCH_PROFILER}
    var
      aModule        : TMod;
      SumClocksFast  : Int64;
      SumClocksSlow  : Int64;
      SumSamplesFast : Int64;
      SumSamplesSlow : Int64;
      SumZips        : Int64;

      function Calc( Clocks, Samples: Int64): string; overload;
      begin
        if Samples = 0
        then Result := '---'
        else Result := Format( '%f', [ Clocks / Samples], AppLocale);
      end;

      function Calc( Clocks, Samples: TSignal): string; overload;
      begin
        if Samples = 0
        then Result := '---'
        else Result := Format( '%f', [ Clocks / Samples], AppLocale);
      end;

    {$ENDIF}
    begin
      Result := TStringList.Create;

      {$IFDEF PATCH_PROFILER}
      SumClocksFast  := 0;
      SumClocksSlow  := 0;
      SumSamplesFast := 0;
      SumSamplesSlow := 0;
      SumZips        := 0;

      Result.Add( '<!doctype html>');
      Result.Add( '<html>');
      Result.Add( '<head><style type="text/css">');
      Result.Add( 'table, td, th { border-style: solid; border-width: 1px; padding: 5px;}');
      Result.Add( '</style></head>');
      Result.Add( '<body><table><tbody>');
      Result.Add( '<tr><td>name</td><td>type</td><td>fast samples</td><td>slow samples</td><td>fast clocks</td><td>slow clocks</td><td>fast clocks / sample</td><td>slow clocks / sample</td></tr>');

      for aModule in FModules
      do begin
        Result.Add( aModule.GetProfileInfoHtml);
        Inc( SumClocksFast , aModule.TotalClocksFast);
        Inc( SumClocksSlow , aModule.TotalClocksSlow);
        Inc( SumSamplesFast, aModule.SamplesFast    );
        Inc( SumSamplesSlow, aModule.SamplesSlow    );
        Inc( SumZips       , aModule.TotalZips      );
      end;

      Result.Add( GetProfileInfoHtml);
      Result.Add(
        Format(
          '<tr><td>%s</td><td>%s</td><td align="right">%d</td><td align="right">%d</td><td align="right">%d</td><td align="right">%d</td><td align="right">%s</td><td align="right">%s</td></tr>',
          [
            'total',
            'average module',
            SumSamplesFast,
            SumSamplesSlow,
            SumClocksFast,
            SumClocksSlow,
            Calc( SumClocksFast, SumSamplesFast),
            Calc( SumClocksSlow, SumSamplesSlow * CONTROL_DEC)
          ],
          AppLocale
        )
      );
      Result.Add(
        Format(
          '<tr><td>%s</td><td>%s</td><td align="right">%d</td><td align="right">%d</td><td align="right">%d</td><td align="right">%s</td><td></td><td></td></tr>',
          [
            'dezippers',
            '&nbsp;',
            SamplesFast,
            SamplesSlow,
            SumZips,
            Calc( SumZips, SamplesFast)
          ],
          AppLocale
        )
      );

      if ModuleCount > 0
      then begin
        Result.Add(
          Format(
            '<tr><td>%s</td><td>%s</td><td align="right">%d</td><td align="right">%d</td><td align="right">%d</td><td align="right">%s</td><td></td><td></td></tr>',
            [
              'dezippers / module',
              '&nbsp;',
              SamplesFast,
              SamplesSlow,
              Round( SumZips / ModuleCount),
              Calc( SumZips, SamplesFast * ModuleCount)
            ],
            AppLocale
          )
        );
      end;

      Result.Add( '<tr><td>name</td><td>type</td><td>fast samples</td><td>slow samples</td><td>fast clocks</td><td>slow clocks</td><td>fast clocks / sample</td><td>slow clocks / sample</td></tr>');
      Result.Add( '</tbody></table></body></html>');
      ClearProfile;
      {$ENDIF}
    end;


    function    TSynthPatch.IsSamePatch( const aPatch: TSynthPatch): Boolean;
    begin
      Result := False;

      if Assigned( aPatch)
      then begin
        Result :=
          SameModules    ( aPatch) and
          SameConnections( aPatch)
      end;
    end;


    function    TSynthPatch.SameModules( const aPatch: TSynthPatch): Boolean;
    var
      i : Integer;
    begin
      Result := False;

      if Assigned( aPatch)
      then begin
        if ModuleCount = aPatch.ModuleCount
        then begin
          Result := True;

          for i := 0 to ModuleCount - 1
          do begin
            if not Module[ i].IsSameModule( aPatch.Module[ i])
            then begin
              Result := False;
              Break;
            end;
          end;
        end;
      end;
    end;


    function    TSynthPatch.SameConnections( const aPatch: TSynthPatch): Boolean;
    begin
      Result := False;

      if Assigned( aPatch)
      then Result := FConnections.IsSameConnections( aPatch.FConnections);
    end;


    procedure   TSynthPatch.AcceptMidi( const aMsg: TMidiMessage); // overide;
    var
      aModule : TMod;
    begin
      if FIsMidi
      then begin
        for aModule in FModules
        do begin
          if aModule.FIsMidi
          then aModule.AcceptMidi( aMsg)
        end;
      end;
    end;


    procedure   TSynthPatch.SendMidi( const aMsg: TMidiMessage); // override;
    begin
      if Assigned( FOnSendMidi)
      then FOnSendMidi( Self, aMsg);
    end;


    procedure   TSynthPatch.AcceptOSC( const aSynthName: string; const aMsg: TOSCPacket); // override;
    var
      aModule : TMod;
    begin
      if IsOSC and Assigned( aMsg)
      then begin
        for aModule in FModules
        do begin
          if aModule.IsOSC
          then begin
            aModule.AcceptOSC( aSynthName, aMsg);

            if aMsg.Matched
            then Break;
          end;
        end;
      end;
    end;


    procedure   TSynthPatch.SendOSCBytes( const aMsg: TBytes); // override;
    begin
      if Assigned( FOnSendOscString)
      then FOnSendOscString( Self, aMsg);
    end;


    procedure   TSynthPatch.SampleRateChanged; // override;
    // Notification for sample rate change, the actual values are in global variables
    // System_Rate for the audio rate and Control_Rate for the control rate.
    // Modules caching sample rate dependent data should recalculate such data in an
    // overriden version of this procedure.
    var
      aModule : TMod;
    begin
      // Notify all sub-modules.
      for aModule in FModules
      do aModule.SampleRateChanged;
    end;


    procedure   TSynthPatch.ClearOverUnderFlows;
    begin
      if Assigned( FInputFifo)
      then FInputFifo.ClearOverUnderFlows;
    end;


    procedure   TSynthPatch.CollectDenormals( const aRange: TSignal; var aMap: TPinMaps);
    var
      i : Integer;
    begin
      SetLength( aMap, ModuleCount);

      for i := 0 to ModuleCount - 1
      do begin
        Module[ i].CollectDenormals( aRange, aMap[ i]);
        aMap[ i].ModuleIndex := i;
      end;
    end;


end.
